class CardboardException(Exception):
    """
    Exception raised for other Cardboard-API exceptions, that don't fall under defined exceptions.

    Also the base exception class.
    """
    pass


class Forbidden(CardboardException):
    """
    Exception raised when there is insufficient permissions.
    """
    pass


class Unauthorized(CardboardException):
    """
    Exception raised when the request lacks valid authentication credentials.
    """
    pass


class NotFound(CardboardException):
    """
    Exception raised when a requested resource is not found.
    """
    pass


class InternalServerError(CardboardException):
    """
    Exception raised when the server encounters an internal error.

    Likely because the API is down.
    """
    pass


class RateLimited(CardboardException):
    """
    Exception raised when the rate limit for requests has been exceeded.
    """
    pass

class BadRequest(CardboardException):
    """
    Exception raised when a bad request is sent.

    Likely invalid data was posted.
    """
    pass


CardboardForbidden = Forbidden
CardboardUnauthorized = Unauthorized
CardboardNotFound = NotFound
CardboardInternalServerError = InternalServerError
CardboardRateLimited = RateLimited
CardboardBadRequest = BadRequest


import requests
from datetime import datetime
from cardboard.Exceptions import *


def _handle_error(response):
    """
    Handles the error responses from API requests.

    Args:
        response (requests.Response): The response object from the API request.

    Raises:
        Forbidden: If the response status code is 403.
        Unauthorized: If the response status code is 401.
        NotFound: If the response status code is 404.
        InternalServerError: If the response status code is 500.
        RateLimited: If the response status code is 429.
        CardboardException: If the response status code does not match any defined error codes.
    """
    if response.status_code == 403:
        raise Forbidden(f"{response.content.decode()}")
    elif response.status_code == 401:
        raise Unauthorized(f"{response.content.decode()}")
    elif response.status_code == 404:
        raise NotFound(f"{response.content.decode()}")
    elif response.status_code == 500:
        raise InternalServerError(f"{response.content.decode()}")
    elif response.status_code == 429:
        raise RateLimited(f"{response.content.decode()}")
    elif response.status_code == 400:
        raise BadRequest(f"{response.content.decode()}")
    elif 200 <= response.status_code < 300:
        return False
    else:
        raise CardboardException(f"{response.content.decode()}")

class Cardboard:
    """
    Base Cardboard class for interacting with the Cardboard API.

    Args:
        client_id (str): Your app's id.
        
        secret (str): Your app's secret.
    """

    def __init__(self, client_id:str, secret:str):
        self.client_id = client_id
        self._baseurl:str = "https://cardboard.ink/api/v1"
        self.secret:str = secret
        self.app_name:str = None
        self._session = requests.Session()
        self._session.headers.update({"Content-Type": "application/x-www-form-urlencoded"})

        def __check_verify(self) -> bool|list:
            """
            Verifies authentication data. ID and Secret.

            Returns:
                bool|list: [True, app_name, app_vanity] if everything is valid, else False.
            """
            resp = requests.post(self._baseurl+'/check', headers={'Content-Type': 'application/x-www-form-urlencoded'}, data={'client_id': self.client_id, 'client_secret': self.secret})
            if resp.status_code != 200:
                return False
            return [True, resp.json()['name'], resp.json()['vanity']]
    
        self.__check_verify = lambda: __check_verify(self)

        self._valid = self.__check_verify()
        if self._valid is False:
            raise CardboardException("Invalid credentials provided. (Is your ID and Secret correct?)")
        self.app_name = self._valid[1]
        self.app_url = f"https://cardboard.ink/a/{self._valid[2]}"

    class UserAlias:
        """
        UserAlias object.

        Represents a user's alias.

        Attributes:
            alias (str|None): The user's alias.
            discriminator (str|None): The user's discriminator.
            name (str): The user's name.
            createdAt (datetime.datetime): When this alias was created.
            _raw_createdAt (str): When this alias was created as an ISO formatted string.
            editedAt (datetime.datetime): When this alias was last updated. (Can be the same time as createdAt)
            _raw_editedAt (str): When this alias was last updated as an ISO formatted string. (Can be the same time as createdAt)
            userId (str): The user's ID.
            gameId (int): The game's ID.
            socialLinkSource (str|None): The social link source for the alias. This is not a URL.
            additionalInfo (dict): Additional info about this alias. This can be an empty dict.
            socialLinkHandle (str|None): The social link to the user for this alias. This is not a URL.
            playerInfo (dict|None): Player info for this alias.
            _raw (dict): Raw alias data directly from the API.
        """
        def __init__(self, data):
            self.alias:str|None = data["alias"]
            self.discriminator:str|None = data["discriminator"]
            self.name:str = data["name"]
            self.createdAt:datetime = datetime.fromisoformat(data["createdAt"])
            self._raw_createdAt:str = data["createdAt"]
            self.editedAt:datetime = datetime.fromisoformat(data["editedAt"])
            self._raw_editedAt:str = data["editedAt"]
            self.userId:str = data["userId"]
            self.gameId:int = data["gameId"]
            self.socialLinkSource:str|None = data["socialLinkSource"]
            self.additionalInfo:dict = data["additionalInfo"]
            self.socialLinkHandle:str|None = data["socialLinkHandle"]
            self.playerInfo:dict|None = data["playerInfo"]
            self._raw:dict = data

    class UserAbout:
        """
        UserAbout object.
        
        Attributes:
            bio (str|None): The user's bio.
            tagline (str|None): The user's tagLine.
            _raw (dict): The raw API data.
        """
        def __init__(self, data):
            self.bio:str|None = data.get("bio")
            self.tagline:str|None = data.get("tagLine")
            self._raw:dict = data

    class UserStatus:
        """
        UserStatus object.
        
        Attributes:
            text (str|None): The user's status text.
            reaction_id (int|None): The associated emoji with the user's stauts.
            _raw_text (dict): The raw API data for text. WARNING: This is highly different from this class.
            _raw_reaction (dict): The raw API data for the reaction. WARNING: This is highly different from this class.
            _raw (dict): The raw API data. WARNING: This is highly different from this class.
        """
        def __init__(self, data):
            try:
                self.text:str|None = data["content"]["document"]["nodes"][0]["nodes"][0]["leaves"][0]["text"]
            except:
                self.text:str|None = None
            self._raw_text:dict = data["content"]
            self.reaction_id:int|None = data["customReactionId"]
            self._raw:dict = data
            self._raw_reaction:dict = data["customReaction"]

    class User:
        """
        User object.
        
        Attributes:
            name (str): The user's username.
            id (str): The user's ID.
            subdomain (str): The user's subdomain.
            aliases (list[UserAlias]): A list of user aliases.
            avatar (str): The link to the user's avatar.
            banner (str): The link to the user's banner.
            status (UserStatus): The user's status.
            moderationStatus (str|None): The user's current moderation status on Guilded.
            aboutInfo (UserAbout): The user's info, including bio and tagline.
            userTransientStatus (str|None): The user's current transient status if applicable.
            _raw (dict): The raw data returned from the API.
        """
        def __init__(self, data):
            self.name:str = data["name"]
            self.id:str = data["id"]
            self.subdomain:str = data["subdomain"]
            self.aliases:list = [Cardboard.UserAlias(data) for data in data["aliases"]]
            self.avatar:str = data["avatar"]
            self.banner:str = data["banner"]
            self.status:Cardboard.UserStatus = Cardboard.UserStatus(data["userStatus"])
            self.moderationStatus:str|None = data["moderationStatus"]
            self.aboutInfo:Cardboard.UserAbout = Cardboard.UserAbout(data["aboutInfo"])
            self.userTransientStatus:str|None = data["userTransientStatus"]
            self._raw:dict = data

    class AuthToken:
        """
        AuthToken object.
        
        Attributes:
            token (str): Your new authorization token.
            refresh_token (str): Your refresh token to get a new authorization token.
            token_type (str): Always "Bearer"
            expires_in (int): How many seconds your new token will last.
            _raw (dict): The raw response from the API.
        """
        def __init__(self, data):
            self.token:str = data["access_token"]
            self.refresh_token:str = data["refresh_token"]
            self.token_type:str = data["token_type"]
            self.expires_in:int = data["expires_in"]
            self._raw:dict = data

    def get_token(self, code:str) -> AuthToken:
        """
        Exchanges your initial code for an authorization token.

        Args:
            code(str): Your initial code.
        """
        grant_type = "authorization_code"
        data = {
            "code": code,
            "client_id": self.client_id,
            "client_secret": self.secret,
            "grant_type": grant_type,
        }
        response = self._session.post(f"{self._baseurl}/token", data=data)
        if response.status_code != 200:
            _handle_error(response)
        return self.AuthToken(response.json())

    def refresh_token(self, refresh_token:str) -> AuthToken:
        """
        Refreshes your authorization token. Your old authorization token and refresh token will no longer work after this.

        Args:
            refresh_token (str): Your refresh token.
        """
        grant_type = "refresh_token"
        data = {
            "refresh_token": refresh_token,
            "client_id": self.client_id,
            "client_secret": self.secret,
            "grant_type": grant_type,
        }
        response = self._session.post(f"{self._baseurl}/token", data=data)
        if response.status_code != 200:
            _handle_error(response)
        return self.AuthToken(response.json())

    def revoke_token(self, token:str) -> None:
        """
        Revokes an authorization token. This also revokes your refresh token associated with this authorization token.

        Args:
            token (str): Your authorization token.
        """
        data = {
            "client_id": self.client_id,
            "client_secret": self.secret,
            "token": token,
        }
        response = self._session.post(f"{self._baseurl}/token/revoke", data=data)
        if response.status_code != 200:
            _handle_error(response)

    def get_user(self, token:str) -> User:
        """
        Fetches user data.

        Args:
            token (str): Your authorization token.
        """
        headers = {"Authorization": f"Bearer {token}"}
        response = self._session.get(f"{self._baseurl}/users/@me", headers=headers)
        if response.status_code != 200:
            _handle_error(response)
        return self.User(response.json())

    def check_token(self, token:str) -> bool:
        """
        Checks whether a token is valid or not.

        Args:
            token (str): Your authorization token.
        """
        data = {
            "token": token
        }
        response = self._session.post(f"{self._baseurl}/token/check", data=data)
        if response.status_code != 200:
            _handle_error(response)
        return response.json()["validity"]


from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='cardboard.py',
    version='0.0.10',
    author='YumYummity',
    author_email='034nop@gmail.com',
    description='API wrapper for https://cardboard.ink/api/v1/',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cardboard-ink/cardboard.py/',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',
    ],
    python_requires='>=3.9',
    install_requires=[
        'requests',
        'aiohttp',
        'flask'
    ],
)


from flask import request, Flask, redirect, session
from functools import wraps
from cardboard import Cardboard, CardboardAsync
import aiohttp, asyncio, time, warnings
import concurrent.futures

class FlaskIntegration:
    """
    A flask integration for Cardboard.

    Args:
        - app: Your Flask app.
        - cardboard: Your Cardboard app. Can be either the normal version or async version; it doesn't matter.
        - session_prefix: The session name where you store the cardboard authentication data. Defaults to "cardboard_"
            - token (eg. "cardboard_token")
            - refresh (eg. "cardboard_prefix")
            - expiry (eg. "cardboard_expiry")
            - rd (eg. "cardboard_rd")
        - login_url: A custom login URL instead of your Cardboard app's default URL.
    """
    def __init__(self, app:Flask, cardboard:Cardboard|CardboardAsync, session_prefix:str="cardboard_", login_url:str=None):
        self.app:Flask = app
        self.cb:Cardboard|CardboardAsync = cardboard
        self.secret = self.cb.secret
        self.client_id = self.cb.client_id
        self.cb = CardboardAsync(client_id=self.client_id, secret=self.secret)
        self.cbsync = concurrent.futures.ThreadPoolExecutor().submit(lambda: Cardboard(client_id=self.client_id, secret=self.secret)).result()
        self.session_token = f"{session_prefix}token"
        self.session_refresh = f"{session_prefix}refresh"
        self.session_expiry = f"{session_prefix}expiry"
        self.session_rd = f"{session_prefix}rd"

        self.app_login_redirect = self.cb.app_url if not login_url else login_url

        if not app.secret_key:
            raise ValueError("Flask app secret key is not set or is empty.")
    
    def _constructAuthToken(self, token, expiry, refresh):
        """
        Constructs an AuthToken class.
        """
        data = {"access_token": token, "refresh_token": refresh, "expires_in": expiry-round(time.time()), "token_type": "Bearer"}
        return self.cb.AuthToken(data=data)

    def getloop(self):
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        return loop

    async def asyncpost(self, url, data=None, headers=None) -> dict|int:
        """
        Makes a post request. Returns json or int.
        """
        async with aiohttp.ClientSession() as cs:
            async with cs.post(url, data=data, headers=headers) as response:
                return await response.json() if response.status == 200 else response.status
    
    def autologin(self, route_function):
        """
        Automatically logs you in with a token, or else redirects you to the app login page. This is async.

        Returns an AuthToken class. You can get the token with token.token

        Will automatically redirect after all processing in the login function is complete if the user came from a route with the loggedin decorator.
        
        Usage:

            ```@app.route('/login')
            @fi.autologin
            def login(token:AuthToken):
                # run code, with token always valid.
            ```
        """
        @wraps(route_function)
        def decorator(*args, **kwargs):
            code = request.args.get('code')
            thetoken = session.get(self.session_token)
            if thetoken and concurrent.futures.ThreadPoolExecutor().submit(lambda: self.cbsync.check_token(thetoken)).result():
                token = thetoken
                if session.get(self.session_expiry) and session.get(self.session_refresh):
                    token = self._constructAuthToken(token, session.get(self.session_expiry), session.get(self.session_refresh))
                elif session.get(self.session_refresh):
                    loop = self.getloop()
                    try:
                        token = loop.run_until_complete(self.cb.refresh_token(session.get(self.session_refresh)))
                    except:
                        return redirect(self.app_login_redirect)
                result = route_function(token, *args, **kwargs)
                if session.get(self.session_rd):
                    a = session.pop(self.session_rd)
                    return redirect(a)
                return result
            if code:
                grant_type = "authorization_code"
                data = {
                    "code": code,
                    "client_id": self.client_id,
                    "client_secret": self.secret,
                    "grant_type": grant_type,
                }
                loop = self.getloop()
                response = loop.run_until_complete(self.asyncpost(f"{self.cb._baseurl}/token", data=data, headers={"Content-Type": "application/x-www-form-urlencoded"}))
                if type(response) == int:
                    return redirect(self.app_login_redirect)
                token = self.cb.AuthToken(response)
            else:
                return redirect(self.app_login_redirect)
            session[self.session_token] = token.token
            session[self.session_expiry] = round(time.time())+token.expires_in
            session[self.session_refresh] = token.refresh_token
            result = route_function(token, *args, **kwargs)
            if session.get(self.session_rd):
                a = session.pop(self.session_rd)
                return redirect(a)
            return result
        return decorator

    def login_code(self, route_function):
        """
        Automatically passes the "code" variable instead of using request.args.get('code').

        Returns None if logged in.

        Does not validate your code; a fake code may be passed.

        ```diff
        - WARNING - Deprecated decorator!
        ```

        Usage:

            ```@app.route('/login')
            @fi.login_code
            def login(code:str|None, *args, **kwargs):
                # your login function.
            ```
        """
        warnings.warn("Deprecated decorator. Please use the autologin decorator instead.", DeprecationWarning)
        @wraps(route_function)
        def decorator(*args, **kwargs):
            thetoken = session.get(self.session_token)
            if thetoken and concurrent.futures.ThreadPoolExecutor().submit(lambda: self.cbsync.check_token(thetoken)).result():
                return route_function(None, *args, **kwargs)
            code = request.args.get('code')
            if not code:
                return redirect(self.app_login_redirect)
            return route_function(code, *args, **kwargs)
        return decorator
    
    def login_autoexchange(self, route_function):
        """
        Automatically exchanges the initial code. This is async.

        Returns None if invalid initial code.

        ```diff
        - WARNING - Deprecated decorator!
        ```

        Usage:

            ```@app.route('/login')
            @fi.login_autoexchange
            def login(token:AuthToken|None, *args, **kwargs):
                # your login function.
            ```
        """
        warnings.warn("Deprecated decorator. Please use the autologin decorator instead.", DeprecationWarning)
        @wraps(route_function)
        def decorator(*args, **kwargs):
            code = request.args.get('code')
            thetoken = session.get(self.session_token)
            if thetoken and concurrent.futures.ThreadPoolExecutor().submit(lambda: self.cbsync.check_token(thetoken)).result():
                token = thetoken
                return route_function(token, *args, **kwargs)
            if code:
                grant_type = "authorization_code"
                data = {
                    "code": code,
                    "client_id": self.client_id,
                    "client_secret": self.secret,
                    "grant_type": grant_type,
                }
                loop = self.getloop()
                response = loop.run_until_complete(self.asyncpost(f"{self.cb._baseurl}/token", data=data, headers={"Content-Type": "application/x-www-form-urlencoded"}))
                if type(response) == int:
                    token = None
                else:
                    token = self.cb.AuthToken(response)
            else:
                token = None
            return route_function(token, *args, **kwargs)
        return decorator
    
    def logged_in(self, route_function):
        """
        Checks if the user is logged in with a valid auth token. Redirects to the app login if not valid.

        After logging in, the user is redirected back to the original URL if the @autologin route is used.
        
        Checking is done in a seperate thread.

        Usage:

            ```@app.route('/dashboard')
            @fi.logged_in
            def dashboard(token:AuthToken, *args, **kwargs):
                # your function
            ```
        """
        @wraps(route_function)
        def decorator(*args, **kwargs):
            thetoken = session.get(self.session_token)
            if thetoken and concurrent.futures.ThreadPoolExecutor().submit(lambda: self.cbsync.check_token(thetoken)).result():
                token = session.get(self.session_token)
                if session.get(self.session_expiry) and session.get(self.session_refresh):
                    token = self._constructAuthToken(token, session.get(self.session_expiry), session.get(self.session_refresh))
                elif session.get(self.session_refresh):
                    loop = self.getloop()
                    try:
                        token = loop.run_until_complete(self.cb.refresh_token(session.get(self.session_refresh)))
                    except:
                        session[self.session_rd] = request.url
                        return redirect(self.app_login_redirect)
                return route_function(token, *args, **kwargs)
            else:
                session[self.session_rd] = request.url
                return redirect(self.app_login_redirect)
        return decorator
    
    def autologout(self, route_function):
        """
        Automatically logs the user out, removing all necessary session data.

        Usage:

            ```@app.route('/logout')
            @fi.autologout
            def logout(*args, **kwargs):
                return redirect(url_for('home'))
            ```
        """
        @wraps(route_function)
        def decorator(*args, **kwargs):
            ot = session.pop(self.session_token, None)
            session.pop(self.session_expiry, None)
            session.pop(self.session_refresh, None)
            session.pop(self.session_rd, None)
            if ot:
                try:
                    concurrent.futures.ThreadPoolExecutor().submit(lambda: self.cbsync.revoke_token(ot)).result()
                except:
                    pass
            return route_function(*args, **kwargs)
        return decorator